# Copyright 2025 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
import httpx
import randomname

from typing import List, Union, Optional, Dict

from qiskit import QuantumCircuit
from qiskit.result import Result
from qiskit.providers import JobV1
from qiskit.providers import JobError, JobTimeoutError, JobStatus

from qiskit_scaleway.versions import USER_AGENT

from qio.core import (
    QuantumProgram,
    QuantumProgramResult,
    QuantumComputationModel,
    QuantumComputationParameters,
    QuantumNoiseModel,
    BackendData,
    ClientData,
)

from scaleway_qaas_client.v1alpha1 import (
    QaaSClient,
    QaaSJobResult,
)


class BaseJob(JobV1):
    def __init__(
        self,
        backend,
        client: QaaSClient,
        circuits: Union[List[QuantumCircuit], QuantumCircuit],
        config: Dict,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(backend, None)
        self._name = name or f"qj-qiskit-{randomname.get_name()}"
        self._client = client
        self._circuits = circuits
        self._config = config
        self._last_progress_message = ""

    @property
    def name(self):
        return self._name

    def status(self) -> JobStatus:
        job = self._client.get_job(self._job_id)

        status_mapping = {
            "running": JobStatus.RUNNING,
            "waiting": JobStatus.QUEUED,
            "completed": JobStatus.DONE,
        }

        if job.progress_message is not None:
            self._last_progress_message = job.progress_message

        return status_mapping.get(job.status, JobStatus.ERROR)

    def submit(self, session_id: str) -> None:
        if self._job_id:
            raise RuntimeError(f"Job already submitted (ID: {self._job_id})")

        options = self._config.copy()
        shots = options.pop("shots")
        memory = options.pop("memory", False)

        programs = list(
            map(lambda c: QuantumProgram.from_qiskit_circuit(c), self._circuits)
        )

        noise_model = options.pop("noise_model", None)
        if noise_model:
            noise_model = QuantumNoiseModel.from_qiskit_aer_noise_model(noise_model)

        backend_data = BackendData(
            name=self.backend().name,
            version=self.backend().version,
            options=options,
        )

        client_data = ClientData(
            user_agent=USER_AGENT,
        )

        computation_model_json = QuantumComputationModel(
            programs=programs,
            backend=backend_data,
            client=client_data,
            noise_model=noise_model,
        ).to_json_str()

        computation_parameters_json = QuantumComputationParameters(
            shots=shots,
            options={
                "memory": memory,
            },
        ).to_json_str()

        model = self._client.create_model(
            payload=computation_model_json,
        )

        if not model:
            raise RuntimeError("Failed to push circuit data")

        self._last_progress_message = ""
        self._job_id = self._client.create_job(
            name=self._name,
            session_id=session_id,
            model_id=model.id,
            parameters=computation_parameters_json,
        ).id

    def result(
        self, timeout: Optional[int] = None, fetch_interval: int = 3
    ) -> Union[Result, List[Result]]:
        if self._job_id == None:
            raise JobError("Job ID error")

        job_results = self._wait_for_result(timeout, fetch_interval)

        program_results = list(
            map(
                lambda r: self._extract_payload_from_response(r).to_qiskit_result(
                    backend_name=self.backend().name,
                    backend_version=self.backend().version,
                    job_id=self._job_id,
                    qobj_id=", ".join(x.name for x in self._circuits),
                    success=r.url is not None or r.result is not None,
                    status=JobStatus.DONE,
                    date=r.created_at,
                ),
                job_results,
            )
        )

        if len(program_results) == 1:
            return program_results[0]

        return program_results

    def _extract_payload_from_response(
        self, job_result: QaaSJobResult
    ) -> QuantumProgramResult:
        result = job_result.result

        if result is None or result == "":
            url = job_result.url

            if url is not None:
                resp = httpx.get(url)
                resp.raise_for_status()
                result = resp.text
            else:
                raise RuntimeError("Got result with empty data and url fields")

        return QuantumProgramResult.from_json_str(result)

    def _wait_for_result(
        self, timeout: Optional[int], fetch_interval: int
    ) -> List[QaaSJobResult]:
        start_time = time.time()

        while True:
            elapsed = time.time() - start_time

            if timeout is not None and elapsed >= timeout:
                raise JobTimeoutError("Timed out waiting for result")

            status = self.status()

            if status == JobStatus.DONE:
                return self._client.list_job_results(self._job_id)

            if status == JobStatus.ERROR:
                raise JobError(f"Job failed: {self._last_progress_message}")

            time.sleep(fetch_interval)



# Copyright 2024 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import re

from numpy.typing import NDArray

from qiskit.primitives.backend_sampler_v2 import (
    BackendSamplerV2,
    _MeasureInfo,
    _samples_to_packed_array,
    QiskitError,
    ResultMemory,
)

from qiskit.primitives.containers import (
    BitArray,
    DataBin,
    SamplerPubResult,
)

_NON_BINARY_CHARS = re.compile(r"[^01]")


class Sampler(BackendSamplerV2):
    def __init__(
        self,
        backend,
        session_id: str,
        options: dict | None = None,
    ):
        if not session_id:
            raise Exception("session_id must be not None")

        if not options:
            options = {}

        if not options.get("run_options"):
            options["run_options"] = {}

        if not options["run_options"].get("session_id"):
            options["run_options"]["session_id"] = session_id

        super().__init__(
            backend=backend,
            options=options,
        )

    def _postprocess_pub(
        self,
        result_memory: list[ResultMemory],
        shots: int,
        shape: tuple[int, ...],
        meas_info: list[_MeasureInfo],
        max_num_bytes: int,
        circuit_metadata: dict,
        meas_level: int | None,
    ) -> SamplerPubResult:
        """Converts the memory data into a sampler pub result

        For level 2 data, the memory data are stored in an array of bit arrays
        with the shape of the pub. For level 1 data, the data are stored in a
        complex numpy array.
        """
        if meas_level == 2 or meas_level is None:
            arrays = {
                item.creg_name: np.zeros(
                    shape + (shots, item.num_bytes), dtype=np.uint8
                )
                for item in meas_info
            }
            memory_array = _memory_array(result_memory, max_num_bytes)

            for samples, index in zip(memory_array, np.ndindex(*shape)):
                for item in meas_info:
                    ary = _samples_to_packed_array(samples, item.num_bits, item.start)
                    arrays[item.creg_name][index] = ary

            meas = {
                item.creg_name: BitArray(arrays[item.creg_name], item.num_bits)
                for item in meas_info
            }
        elif meas_level == 1:
            raw = np.array(result_memory)
            cplx = raw[..., 0] + 1j * raw[..., 1]
            cplx = np.reshape(cplx, (*shape, *cplx.shape[1:]))
            meas = {item.creg_name: cplx for item in meas_info}
        else:
            raise QiskitError(f"Unsupported meas_level: {meas_level}")
        return SamplerPubResult(
            DataBin(**meas, shape=shape),
            metadata={"shots": shots, "circuit_metadata": circuit_metadata},
        )


def _memory_array(results: list[list[str]], num_bytes: int) -> NDArray[np.uint8]:
    """Converts the memory data into an array in an unpacked way."""
    lst = []
    # Heuristic: check only the first result format
    if len(results) > 0 and len(results[0]) > 0:
        base = 16 if _NON_BINARY_CHARS.search(results[0][0]) else 2

    for memory in results:
        if num_bytes > 0:
            data = b"".join(int(i, base).to_bytes(num_bytes, "big") for i in memory)
            data = np.frombuffer(data, dtype=np.uint8).reshape(-1, num_bytes)
        else:
            # no measure in a circuit
            data = np.zeros((len(memory), num_bytes), dtype=np.uint8)
        lst.append(data)
    ary = np.asarray(lst)
    return np.unpackbits(ary, axis=-1, bitorder="big")



# Copyright 2024 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import (
    List,
    Union,
    Dict,
)

from qiskit import QuantumCircuit
from qiskit.providers import JobError, JobStatus
from qiskit.result import Result
from qiskit.transpiler.passes import RemoveBarriers

from qiskit_scaleway.versions import USER_AGENT
from qiskit_scaleway.backends import BaseJob

from qio.core import (
    QuantumProgram,
    QuantumProgramSerializationFormat,
    QuantumProgramResult,
    QuantumComputationModel,
    QuantumComputationParameters,
    BackendData,
    ClientData,
)

from scaleway_qaas_client.v1alpha1 import QaaSClient


class QsimJob(BaseJob):
    def __init__(
        self,
        name: str,
        backend,
        client: QaaSClient,
        circuits: Union[List[QuantumCircuit], QuantumCircuit],
        config: Dict,
    ) -> None:
        super().__init__(
            name=name, backend=backend, client=client, config=config, circuits=circuits
        )

    def submit(self, session_id: str) -> None:
        if self._job_id:
            raise RuntimeError(f"Job already submitted (ID: {self._job_id})")

        options = self._config.copy()

        # Note 1: Barriers are only visual elements
        # Barriers are not managed by Cirq deserialization
        # Note 2: Qsim can only handle one circuit at a time
        circuit = RemoveBarriers()(self._circuits[0])

        programs = [
            QuantumProgram.from_qiskit_circuit(
                circuit, QuantumProgramSerializationFormat.QASM_V2
            )
        ]

        options.pop("circuit_memoization_size")
        shots = options.pop("shots")

        backend_data = BackendData(
            name=self.backend().name,
            version=self.backend().version,
            options=options,
        )

        client_data = ClientData(
            user_agent=USER_AGENT,
        )

        computation_model_json = QuantumComputationModel(
            programs=programs,
            backend=backend_data,
            client=client_data,
        ).to_json_str()

        computation_parameters_json = QuantumComputationParameters(
            shots=shots,
        ).to_json_str()

        model = self._client.create_model(
            payload=computation_model_json,
        )

        if not model:
            raise RuntimeError("Failed to push circuit data")

        self._job_id = self._client.create_job(
            name=self._name,
            session_id=session_id,
            model_id=model.id,
            parameters=computation_parameters_json,
        ).id

    def __to_qiskit_result(self, program_result: QuantumProgramResult) -> Result:
        status = self.status()

        return program_result.to_qiskit_result(
            backend_name=self.backend().name,
            backend_version=self.backend().version,
            job_id=self._job_id,
            qobj_id=", ".join(x.name for x in self._circuits),
            success=status == JobStatus.DONE,
            status=JobStatus.DONE,
        )

    def result(
        self,
        timeout=None,
        fetch_interval: int = 3,
    ) -> Union[Result, List[Result]]:
        if self._job_id == None:
            raise JobError("Job ID error")
        job_results = self._wait_for_result(timeout, fetch_interval)

        program_results = list(
            map(
                lambda r: self.__to_qiskit_result(
                    self._extract_payload_from_response(r)
                ),
                job_results,
            )
        )

        if len(program_results) == 1:
            return program_results[0]

        return program_results



# Copyright 2026 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import (
    List,
    Union,
    Dict,
)

from qiskit import QuantumCircuit

from qiskit_scaleway.versions import USER_AGENT
from qiskit_scaleway.backends import BaseJob
from qiskit.transpiler.passes import RemoveBarriers

from qio.core import (
    QuantumProgram,
    QuantumProgramSerializationFormat,
    QuantumComputationModel,
    QuantumComputationParameters,
    BackendData,
    ClientData,
)

from scaleway_qaas_client.v1alpha1 import QaaSClient


class CudaqJob(BaseJob):
    def __init__(
        self,
        name: str,
        backend,
        client: QaaSClient,
        circuits: Union[List[QuantumCircuit], QuantumCircuit],
        config: Dict,
    ) -> None:
        super().__init__(
            name=name, backend=backend, client=client, config=config, circuits=circuits
        )

    def submit(self, session_id: str) -> None:
        if self._job_id:
            raise RuntimeError(f"Job already submitted (ID: {self._job_id})")

        options = self._config.copy()

        circuit = RemoveBarriers()(self._circuits[0])

        programs = [
            QuantumProgram.from_qiskit_circuit(
                circuit, QuantumProgramSerializationFormat.QASM_V3
            )
        ]

        # Retrieve run options
        shots = options.pop("shots")
        cudaq_option = {"cudaq_target_option": options.pop("option", "")}

        backend_data = BackendData(
            name=self.backend().name,
            version=self.backend().version,
            options=options,
        )

        client_data = ClientData(
            user_agent=USER_AGENT,
        )

        computation_model_json = QuantumComputationModel(
            programs=programs,
            backend=backend_data,
            client=client_data,
        ).to_json_str()

        computation_parameters_json = QuantumComputationParameters(
            shots=shots,
            options=cudaq_option,
        ).to_json_str()

        model = self._client.create_model(
            payload=computation_model_json,
        )

        if not model:
            raise RuntimeError("Failed to push circuit data")

        self._job_id = self._client.create_job(
            name=self._name,
            session_id=session_id,
            model_id=model.id,
            parameters=computation_parameters_json,
        ).id
